@@ -1,5 +1,6 @@
 /*
- * Copyright (C) 2022 The Phosh Developers
+ * Copyright (C) 2022-2024 The Phosh Developers
+ *               2025 Phosh.mobi e.V.
  *
  * SPDX-License-Identifier: GPL-3.0-or-later
  *
@@ -19,6 +20,8 @@
 #include "pos-osk-key.h"
 #include "pos-osk-widget.h"
 
+#include "gmobile.h"
+
 #include <json-glib/json-glib.h>
 #include <pango/pangocairo.h>
 
@@ -32,8 +35,7 @@
 
 #define MINIMUM_WIDTH 360
 
-#define KEY_REPEAT_DELAY 700
-#define KEY_REPEAT_INTERVAL 50
+#define EVENT_HISTORY_THRESHOLD_MS 150
 
 enum {
   OSK_KEY_DOWN,
@@ -57,6 +59,12 @@ enum {
 };
 static GParamSpec *props[PROP_LAST_PROP];
 
+typedef enum {
+  CURSOR_DRAG_STARTING,
+  CURSOR_DRAG_HORIZ,
+  CURSOR_DRAG_VERT,
+} cursor_drag_t;
+
 /**
  * PosOskWidgetRow:
  * @width: number in key units
@@ -153,9 +161,89 @@ struct _PosOskWidget {
   /* Cursor movement */
   GtkGesture          *cursor_drag;
   double               last_x, last_y;
+  GArray              *event_history;
+  cursor_drag_t        drag_type;
 };
 G_DEFINE_TYPE (PosOskWidget, pos_osk_widget, GTK_TYPE_DRAWING_AREA)
 
+typedef struct {
+  double delta_x, delta_y;
+  guint32 time;
+} EventHistoryRecord;
+
+
+static void
+pos_osk_widget_calculate_drag_velocity (PosOskWidget *self, double *v_x, double *v_y)
+{
+  gdouble total_delta_x = 0, total_delta_y = 0;
+  guint32 first_time = 0, last_time = 0;
+  guint i;
+
+  for (i = 0; i < self->event_history->len; i++) {
+    EventHistoryRecord *r =
+      &g_array_index (self->event_history, EventHistoryRecord, i);
+
+    if (i == 0)
+      first_time = r->time;
+    else {
+      total_delta_x += r->delta_x;
+      total_delta_y += r->delta_y;
+    }
+
+    last_time = r->time;
+  }
+
+  if (first_time == last_time) {
+    if (v_x)
+      *v_x = total_delta_x / (last_time - first_time);
+
+    if (v_y)
+      *v_y = total_delta_y / (last_time - first_time);
+  }
+
+  if (v_x)
+    *v_x = total_delta_x / (last_time - first_time);
+
+  if (v_y)
+    *v_y = total_delta_y / (last_time - first_time);
+}
+
+
+static void
+pos_osk_widget_trim_event_history (PosOskWidget *self)
+{
+  g_autoptr (GdkEvent) event = gtk_get_current_event ();
+  guint32 threshold_time = gdk_event_get_time (event) - EVENT_HISTORY_THRESHOLD_MS;
+  guint i;
+
+  for (i = 0; i < self->event_history->len; i++) {
+    guint32 time = g_array_index (self->event_history,
+                                  EventHistoryRecord, i).time;
+
+    if (time >= threshold_time)
+      break;
+  }
+
+  if (i > 0)
+    g_array_remove_range (self->event_history, 0, i);
+}
+
+
+static void
+pos_osk_widget_append_to_event_history (PosOskWidget *self, double delta_x, double delta_y)
+{
+  g_autoptr (GdkEvent) event = gtk_get_current_event ();
+  EventHistoryRecord record;
+
+  pos_osk_widget_trim_event_history (self);
+
+  record.delta_x = delta_x;
+  record.delta_y = delta_y;
+  record.time = gdk_event_get_time (event);
+
+  g_array_append_val (self->event_history, record);
+}
+
 
 static void
 on_drag_begin (PosOskWidget *self,
@@ -164,33 +252,60 @@ on_drag_begin (PosOskWidget *self,
 {
   if (self->mode != POS_OSK_WIDGET_MODE_CURSOR)
     return;
+
+  self->drag_type = CURSOR_DRAG_STARTING;
+  self->last_x = start_x;
+  self->last_y = start_y;
 }
 
 #define KEY_DIST_X 5
 #define KEY_DIST_Y 10
+#define V_X_MIN 1.4
+#define V_Y_MIN 2.8
+
+#define CAN_DRAG_HORIZ(self)                                            \
+  (self->drag_type == CURSOR_DRAG_STARTING || self->drag_type == CURSOR_DRAG_HORIZ)
+#define CAN_DRAG_VERT(self)                                             \
+  (self->drag_type == CURSOR_DRAG_STARTING || self->drag_type == CURSOR_DRAG_VERT)
+#define DRAG_TYPE(self)                                                 \
+  (self->drag_type == CURSOR_DRAG_STARTING ? "starting"                 \
+   : self->drag_type == CURSOR_DRAG_HORIZ  ? "horiz"                    \
+   : "vert")
+
 
 static void
-on_drag_update (PosOskWidget *self,
-                double        off_x,
-                double        off_y)
+on_drag_update (PosOskWidget *self, double off_x, double off_y)
 {
   const char *symbol = NULL;
-  double delta_x, delta_y;
+  double delta_x, delta_y, v_x, v_y;
 
   if (self->mode != POS_OSK_WIDGET_MODE_CURSOR)
     return;
 
-  g_debug ("%s: %f, %f", __func__, off_x, off_y);
+  g_debug ("%s: (%f,%f) %s", __func__, off_x, off_y, DRAG_TYPE (self));
 
   delta_x = self->last_x - off_x;
   delta_y = self->last_y - off_y;
+  pos_osk_widget_append_to_event_history (self, delta_x, delta_y);
 
-  if (ABS (delta_x) > KEY_DIST_X) {
+  pos_osk_widget_calculate_drag_velocity (self, &v_x, &v_y);
+
+  if (CAN_DRAG_HORIZ (self) && ABS (v_x) < V_X_MIN && ABS (v_y) > V_Y_MIN) {
+    g_debug ("Switching to vert: %f %f", v_x, v_y);
+    self->drag_type = CURSOR_DRAG_VERT;
+  } else if (CAN_DRAG_VERT (self) && ABS (v_y) < V_Y_MIN && ABS (v_x) > V_X_MIN) {
+    g_debug ("Switching to horiz: %f %f", v_x, v_y);
+    self->drag_type = CURSOR_DRAG_HORIZ;
+  }
+
+  if (ABS (delta_x) > KEY_DIST_X && CAN_DRAG_HORIZ (self)) {
     symbol =  delta_x > 0 ? POS_OSK_SYMBOL_LEFT : POS_OSK_SYMBOL_RIGHT;
     self->last_x = off_x;
-  } else if (ABS (delta_y) > KEY_DIST_Y) {
+    self->drag_type = CURSOR_DRAG_HORIZ;
+  } else if (ABS (delta_y) > KEY_DIST_Y && CAN_DRAG_VERT (self)) {
     symbol =  delta_y > 0 ? POS_OSK_SYMBOL_UP : POS_OSK_SYMBOL_DOWN;
     self->last_y = off_y;
+    self->drag_type = CURSOR_DRAG_VERT;
   }
 
   if (symbol)
@@ -205,16 +320,14 @@ on_drag_end (PosOskWidget *self)
     return;
 
   pos_osk_widget_set_mode (self, POS_OSK_WIDGET_MODE_KEYBOARD);
+  g_array_remove_range (self->event_history, 0, self->event_history->len);
 }
 
 
 static void
 on_drag_cancel (PosOskWidget *self)
 {
-  if (self->mode != POS_OSK_WIDGET_MODE_CURSOR)
-    return;
-
-  pos_osk_widget_set_mode (self, POS_OSK_WIDGET_MODE_KEYBOARD);
+  on_drag_end (self);
 }
 
 
@@ -409,6 +522,7 @@ get_key (PosOskWidget *self, const char *symbol, GStrv symbols, const char *labe
                          "symbols", symbols,
                          "width", 2.0,
                          "expand", TRUE,
+                         "style", "space",
                          NULL);
   }
   return g_object_new (POS_TYPE_OSK_KEY,
@@ -843,33 +957,6 @@ pos_osk_widget_locate_key (PosOskWidget *self, double x, double y)
 }
 
 
-static gboolean
-on_key_repeat (gpointer data)
-{
-  PosOskWidget *self = POS_OSK_WIDGET (data);
-
-  g_return_val_if_fail (self->current, G_SOURCE_REMOVE);
-
-  g_signal_emit (self, signals[OSK_KEY_DOWN], 0, pos_osk_key_get_symbol (self->current));
-  g_signal_emit (self, signals[OSK_KEY_UP], 0, pos_osk_key_get_symbol (self->current));
-  g_signal_emit (self, signals[OSK_KEY_SYMBOL], 0, pos_osk_key_get_symbol (self->current));
-
-  return G_SOURCE_CONTINUE;
-}
-
-
-static gboolean
-on_repeat_timeout (gpointer data)
-{
-  PosOskWidget *self = POS_OSK_WIDGET (data);
-
-  self->repeat_id = g_timeout_add (KEY_REPEAT_INTERVAL, on_key_repeat, self);
-  g_source_set_name_by_id (self->repeat_id, "[pos-key-repeat]");
-
-  return G_SOURCE_REMOVE;
-}
-
-
 static void
 key_repeat_cancel (PosOskWidget *self)
 {
@@ -919,11 +1006,6 @@ pos_osk_widget_key_press (PosOskWidget *self, double x, double y)
   }
   pos_osk_widget_key_press_action (self, key);
 
-  if (pos_osk_key_get_use (key) == POS_OSK_KEY_USE_DELETE) {
-    self->repeat_id = g_timeout_add (KEY_REPEAT_DELAY, on_repeat_timeout, self);
-    g_source_set_name_by_id (self->repeat_id, "[pos-key-repeat-timeout]");
-  }
-
   return GDK_EVENT_PROPAGATE;
 }
 
@@ -986,8 +1068,8 @@ pos_osk_widget_key_release_action (PosOskWidget *self, PosOskKey *key)
     G_GNUC_FALLTHROUGH;
   case POS_OSK_KEY_USE_DELETE:
     pos_osk_widget_set_key_pressed (self, self->current, FALSE);
-    g_signal_emit (self, signals[OSK_KEY_UP], 0, pos_osk_key_get_symbol (key));
     g_signal_emit (self, signals[OSK_KEY_SYMBOL], 0, pos_osk_key_get_symbol (key));
+    g_signal_emit (self, signals[OSK_KEY_UP], 0, pos_osk_key_get_symbol (key));
     switch_layer (self, key);
     break;
 
@@ -1142,6 +1224,7 @@ on_symbol_selected (PosOskWidget *self, const char *symbol)
 
   g_signal_emit (self, signals[OSK_KEY_DOWN], 0, symbol);
   g_signal_emit (self, signals[OSK_KEY_SYMBOL], 0, symbol);
+  g_signal_emit (self, signals[OSK_KEY_UP], 0, symbol);
   g_clear_pointer (&self->char_popup, phosh_cp_widget_destroy);
 }
 
@@ -1280,9 +1363,10 @@ render_hint (cairo_t *cr, GtkStyleContext *context, const char *hint, const GdkR
   int hint_margin = 1;
   float hint_scale = 0.75;
 
-  gtk_style_context_set_state (context, state);
   cairo_save (cr);
 
+  gtk_style_context_add_class (context, "hint");
+
   gtk_style_context_get (context, state, "font", &font, NULL);
   size = pango_font_description_get_size (font);
   pango_font_description_set_size (font, hint_scale * size);
@@ -1311,7 +1395,7 @@ render_hint (cairo_t *cr, GtkStyleContext *context, const char *hint, const GdkR
   pango_cairo_show_layout (cr, layout);
 
   cairo_restore (cr);
-  gtk_style_context_set_state (context, GTK_STATE_FLAG_NORMAL);
+  gtk_style_context_remove_class (context, "hint");
 }
 
 
@@ -1487,6 +1571,7 @@ pos_osk_widget_finalize (GObject *object)
   g_clear_pointer (&self->region, g_free);
   g_clear_pointer (&self->layout_id, g_free);
   g_ptr_array_free (self->symbols, TRUE);
+  g_clear_pointer (&self->event_history, g_array_unref);
 
   G_OBJECT_CLASS (pos_osk_widget_parent_class)->finalize (object);
 }
@@ -1694,6 +1779,7 @@ pos_osk_widget_init (PosOskWidget *self)
   self->layer = POS_OSK_WIDGET_LAYER_NORMAL;
   self->symbols = g_ptr_array_new ();
   self->key_height = POS_OSK_WIDGET_KEY_HEIGHT_DEFAULT;
+  self->event_history = g_array_new (FALSE, FALSE, sizeof (EventHistoryRecord));
 
   gtk_widget_add_events (GTK_WIDGET (self), GDK_BUTTON_PRESS_MASK |
                          GDK_BUTTON_RELEASE_MASK |
@@ -1814,7 +1900,7 @@ parse_lang (PosOskWidget *self, const char *layout, const char *variant)
 
   /* Keyboard layout has language (`en`), region is from layout (`us`) */
   self->lang = g_strdup (self->layout.locale);
-  if (STR_IS_NULL_OR_EMPTY (variant)) {
+  if (gm_str_is_null_or_empty (variant)) {
     self->region = g_strdup (layout);
     return;
   }
@@ -1868,7 +1954,7 @@ pos_osk_widget_set_layout (PosOskWidget *self,
   g_free (self->layout_id);
   self->layout_id = g_strdup (layout_id);
 
-  if (!STR_IS_NULL_OR_EMPTY (variant))
+  if (!gm_str_is_null_or_empty (variant))
     path = g_strdup_printf ("/mobi/phosh/stevia/layouts/%s+%s.json", layout, variant);
   else
     path = g_strdup_printf ("/mobi/phosh/stevia/layouts/%s.json", layout);
